<?php

namespace Drupal\lupus_ce_renderer;

use drunomics\ServiceUtils\Core\Routing\CurrentRouteMatchTrait;
use drunomics\ServiceUtils\Symfony\HttpFoundation\RequestStackTrait;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\TitleResolverInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\PageCache\ResponsePolicy\KillSwitch;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\custom_elements\CustomElement;
use Drupal\custom_elements\CustomElementGeneratorTrait;
use Drupal\custom_elements\CustomElementNormalizerTrait;
use Drupal\lupus_ce_renderer\Cache\CustomElementsJsonResponse;
use Drupal\metatag\MetatagManagerInterface;
use Drupal\Core\Block\BlockManager;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Site\Settings;
use Symfony\Component\HttpKernel\Exception\NotAcceptableHttpException;

/**
 * Renders main content only.
 */
class CustomElementsRenderer {

  use CustomElementGeneratorTrait;
  use CustomElementNormalizerTrait;
  use CurrentRouteMatchTrait;
  use RequestStackTrait;

  /**
   * Custom elements content formats.
   */
  const CONTENT_FORMAT_MARKUP = 'markup';
  const CONTENT_FORMAT_JSON   = 'json';

  /**
   * The title resolver.
   *
   * @var \Drupal\Core\Controller\TitleResolverInterface
   */
  protected $titleResolver;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The renderer configuration array.
   *
   * @var array
   * @see sites/default/default.services.yml
   */
  protected $rendererConfig;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The messenger interface.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The block plugin manager.
   *
   * @var \Drupal\Core\Block\BlockManager
   */
  protected $blockManager;

  /**
   * The drupal settings.
   *
   * @var \Drupal\Core\Site\Settings
   */
  protected $settings;

  /**
   * The metatag manager.
   *
   * @var \Drupal\metatag\MetatagManagerInterface
   */
  protected $metatagManager;

  /**
   * The kill switch.
   *
   * @var \Drupal\Core\PageCache\ResponsePolicy\KillSwitch
   */
  protected $killSwitch;

  /**
   * Constructs the renderer.
   *
   * @param \Drupal\Core\Controller\TitleResolverInterface $title_resolver
   *   The title resolver.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   * @param array $renderer_config
   *   The renderer configuration array.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger interface.
   * @param \Drupal\Core\Block\BlockManager $blockManager
   *   The block manager.
   * @param \Drupal\Core\Site\Settings $settings
   *   The drupal settings.
   * @param \Drupal\metatag\MetatagManagerInterface $metatagManager
   *   The module handler.
   * @param \Drupal\Core\PageCache\ResponsePolicy\KillSwitch $killSwitch
   *   The kill switch.
   */
  public function __construct(TitleResolverInterface $title_resolver, RendererInterface $renderer, array $renderer_config, ModuleHandlerInterface $module_handler, MessengerInterface $messenger, BlockManager $blockManager, Settings $settings, MetatagManagerInterface $metatagManager, KillSwitch $killSwitch) {
    $this->titleResolver = $title_resolver;
    $this->renderer = $renderer;
    $this->rendererConfig = $renderer_config;
    $this->moduleHandler = $module_handler;
    $this->messenger = $messenger;
    $this->blockManager = $blockManager;
    $this->settings = $settings;
    $this->metatagManager = $metatagManager;
    $this->killSwitch = $killSwitch;
  }

  /**
   * {@inheritdoc}
   */
  public function renderResponse(CustomElement $custom_element, $format = 'markup') {
    $custom_element->addCacheContexts($this->rendererConfig['required_cache_contexts']);
    $bubbleable_metadata = BubbleableMetadata::createFromObject($custom_element);

    if ($format == static::CONTENT_FORMAT_MARKUP) {
      $build = $custom_element->toRenderArray();
      $content = $this->renderer->renderRoot($build);
      $bubbleable_metadata = $bubbleable_metadata
        ->merge(BubbleableMetadata::createFromRenderArray($build));
    }
    else if ($format == static::CONTENT_FORMAT_JSON) {
      $content = $this->getCustomElementNormalizer()->normalize($custom_element, NULL, ['cache_metadata' => $bubbleable_metadata]);
    }
    else {
      throw new \LogicException('Unsupported content format given.');
    }

    $route_match = $this->getCurrentRouteMatch();
    $request = $this->getCurrentRequest();
    $title = $this->titleResolver->getTitle($request, $route_match->getRouteObject());
    $messages = $this->getMessages();
    $breadcrumbs = $this->getBreadcrumbs($bubbleable_metadata);
    $metatags = $this->getMetatags($route_match);

    $data = [
      'title' => $title,
      'messages' => $messages,
      'breadcrumbs' => $breadcrumbs['data'],
      'metatags' => $metatags,
      'content_format' => $format,
      'content' => $content ?: FALSE,
    ];

    if ($format == static::CONTENT_FORMAT_MARKUP) {
      $data['breadcrumbs_html'] = $breadcrumbs['markup'];
      $data['page'] = 'layout--default';
    }

    $this->moduleHandler->alter('lupus_ce_renderer_response', $data, $bubbleable_metadata, $request);

    $response = new CustomElementsJsonResponse();
    $response->setData($data);
    $response->addCacheableDependency($bubbleable_metadata);
    return $response;
  }

  /**
   * Get additional dynamic content.
   *
   * Content will be added to response on late stage. It will not be cached and
   * will not impact dynamic page cache.
   *
   * @return array[]
   *   Data array.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function getDynamicContent() {
    return [
      'local_tasks' => $this->getLocalTasks(),
    ];
  }

  /**
   * Get drupal messages.
   *
   * @return array
   */
  private function getMessages() {
    $messages = $this->messenger->all();
    if (!empty($messages)) {
      $this->killSwitch->trigger();
    }
    if (isset($messages['status'])) {
      $success_messages = $messages['status'];
      unset($messages['status']);
      $messages['success'] = $success_messages;
    }
    $this->messenger->deleteAll();
    return $messages;
  }

  /**
   * Get breadcrumbs data and markup.
   *
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleable_metadata
   *   The bubbleable metatada.
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function getBreadcrumbs(BubbleableMetadata &$bubbleable_metadata) {
    try {
      $breadcrumbs = [];
      $breadcrumb_block = $this->blockManager->createInstance('system_breadcrumb_block', []);
      $breadcrumb_render = $breadcrumb_block->build();
      if (!empty($breadcrumb_render['#links'])) {
        /** @var \Drupal\Core\Link $crumb */
        foreach ($breadcrumb_render['#links'] as $crumb) {
          $text = $crumb->getText();
          if (is_array($crumb->getText()) && isset($text['#markup'])) {
            $text = strip_tags($text['#markup']);
          }
          $breadcrumbs[] = [
            'frontpage' => $crumb->getUrl()->getRouteName() == '<front>',
            'url' => $crumb->getUrl()->toString(),
            'label' => (string) $text,
          ];
        }
        $bubbleable_metadata = BubbleableMetadata::createFromRenderArray($breadcrumb_render)
          ->merge($bubbleable_metadata);
      }

      return [
        'data' => $breadcrumbs,
        'markup' => $this->renderer->renderRoot($breadcrumb_render),
      ];
    }
    // On not-found routes breadcrumbs might not be derivable, thus set them
    // empty.
    catch (NotAcceptableHttpException $exception) {
      return [
        'data' => [],
        'markup' => '',
      ];
    }
  }

  /**
   * Get local-tasks data.
   *
   * @return array
   *   Array of local-task links.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function getLocalTasks() {
    $local_tasks = [];
    $local_tasks_block = $this->blockManager->createInstance('local_tasks_block', []);
    $local_tasks_render = $local_tasks_block->build();
    if (isset($local_tasks_render['#primary'])) {
      $local_tasks['primary'] = $this->prepareLocalTaskLinks($local_tasks_render['#primary']);
    }
    if (isset($local_tasks_render['#secondary'])) {
      $local_tasks['secondary'] = $this->prepareLocalTaskLinks($local_tasks_render['#secondary']);
    }
    return $local_tasks;
  }

  /**
   * Prepare local task links data.
   *
   * @param array $render_links
   *   Renderable links array.
   *
   * @return array
   *   Array of local-task links.
   */
  private function prepareLocalTaskLinks(array $render_links) {
    $local_tasks = [];
    if (!empty($render_links)) {
      foreach ($render_links as $render_link) {
        /** @var $access \Drupal\Core\Access\AccessResult */
        $access = $render_link['#access'] ?? FALSE;
        // Check access to links.
        if ($access instanceof AccessResult && $access->isAllowed() || $access === TRUE) {
          $url = $render_link['#link']['url'] ?? FALSE;
          $title = $render_link['#link']['title'] ?? FALSE;
          if ($url && $title) {
            $local_tasks[] = [
              'url' => $url->toString(),
              'label' => $title,
              'active' => $render_link['#active'] ?? FALSE,
            ];
          }
        }
      }
    }
    return $local_tasks;
  }

  /**
   * Get metatags.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match, for context.
   *
   * @return array
   */
  private function getMetatags(RouteMatchInterface $route_match) {
    $metatags = metatag_generate_entity_metatags($route_match->getParameter('node'));
    $prepared_metatags = [
      'meta' => [],
      'link' => [],
    ];
    $blacklisted_metatags = $this->settings->get('blacklisted_metatags') ?? [];
    foreach ($metatags as $metatag_values) {
      $blacklisted = FALSE;
      if (isset($blacklisted_metatags[$metatag_values['#tag']])) {
        foreach ($metatag_values['#attributes'] as $value) {
          if (in_array($value, $blacklisted_metatags[$metatag_values['#tag']])) {
            $blacklisted = TRUE;
          }
        }
      }
      if (!$blacklisted) {
        $prepared_metatags[$metatag_values['#tag']][] = $metatag_values['#attributes'];
      }
    }
    return $prepared_metatags;
  }

}
